"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import re

from goose3.extractors import BaseExtractor

TITLE_SPLITTERS = ["|", "-", "»", ":"]


class TitleExtractor(BaseExtractor):
    def clean_title(self, title):
        """Clean title with the use of og:site_name in this case try to get rid of site name and use
        TITLE_SPLITTERS to reformat title"""

        # try to remove the domain from url
        if self.article.domain:
            pattern = re.compile(self.article.domain, re.IGNORECASE)
            title = pattern.sub("", title).strip()

        schema = self.article.schema
        site_name = ""
        # check if we have the site name in opengraph data
        if "site_name" in self.article.opengraph and self.article.opengraph["site_name"] != title:
            site_name = self.article.opengraph["site_name"]
        elif schema and schema.get("publisher") and schema["publisher"].get("name"):
            site_name = self.article.schema["publisher"]["name"]

        # if there is a sperator, speratate and check if site name is present
        seps = [s for s in TITLE_SPLITTERS if s in title]
        if seps:
            # NOTE: Perhaps it should only be removed if, once seperated it is "by itself"
            title = title.replace(site_name, "").strip()

        # split the title in words
        # TechCrunch | my wonderfull article
        # my wonderfull article | TechCrunch
        title_words = title.split()

        # check if first letter is in TITLE_SPLITTERS
        # if so remove it
        if title_words and title_words[0] in TITLE_SPLITTERS:
            title_words.pop(0)

        # check for a title that is empty or consists of only a
        # title splitter to avoid a IndexError below
        if not title_words:
            return ""

        # check if last letter is in TITLE_SPLITTERS
        # if so remove it
        if title_words[-1] in TITLE_SPLITTERS:
            title_words.pop(-1)

        # rebuild the title
        title = " ".join(title_words).strip()

        return title

    def get_title(self):
        """Fetch the article title and analyze it"""
        title = ""

        # rely on opengraph in case we have the data
        if "title" in self.article.opengraph:
            return self.clean_title(self.article.opengraph["title"])
        if self.article.schema and "headline" in self.article.schema:
            return self.clean_title(self.article.schema["headline"])

        # try to fetch the meta headline
        meta_headline = self.parser.get_elements_by_tag(self.article.doc, tag="meta", attr="name", value="headline")
        if meta_headline is not None and len(meta_headline) > 0:
            title = self.parser.get_attribute(meta_headline[0], "content")
            return self.clean_title(title)

        # otherwise use the title meta
        title_element = self.parser.get_elements_by_tag(self.article.doc, tag="title")
        if title_element is not None and len(title_element) > 0:
            title = self.parser.get_text(title_element[0])
            return self.clean_title(title)

        return title

    def extract(self):
        return self.get_title()



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import re
from urllib.parse import urljoin, urlparse

from goose3.extractors import BaseExtractor
from goose3.image import Image
from goose3.utils import FileHelper
from goose3.utils.images import ImageUtils

KNOWN_IMG_DOM_NAMES = [
    "yn-story-related-media",
    "cnn_strylccimg300cntr",
    "big_photo",
    "ap-smallphoto-a",
]


class DepthTraversal:
    def __init__(self, node, parent_depth, sibling_depth):
        self.node = node
        self.parent_depth = parent_depth
        self.sibling_depth = sibling_depth


class ImageExtractor(BaseExtractor):
    def __init__(self, fetcher, config, article):
        super().__init__(config, article)

        self.fetcher = fetcher
        self.custom_site_mapping = {}

        self.load_customesite_mapping()

        # What's the minimum bytes for an image we'd accept is
        self.images_min_bytes = 4000

        # this lists all the known bad button names that we have
        self.badimages_names_re = re.compile(
            ".html|.gif|.ico|button|twitter.jpg|facebook.jpg|ap_buy_photo"
            "|digg.jpg|digg.png|delicious.png|facebook.png|reddit.jpg"
            "|doubleclick|diggthis|diggThis|adserver|/ads/|ec.atdmt.com"
            "|mediaplex.com|adsatt|view.atdmt"
        )

    def get_best_image(self, top_node):
        image = self.check_known_elements()
        if image:
            return image

        image = self.check_large_images(top_node, 0, 0)
        if image:
            return image

        image = self.check_meta_tag()
        if image:
            return image
        return Image()

    def check_meta_tag(self):
        # check link tag
        image = self.check_link_tag()
        if image:
            return image

        # check opengraph tag
        return self.check_known_schemas()

    def check_large_images(self, node, parent_depth_level, sibling_depth_level):
        """although slow the best way to determine the best image is to download them and check the actual dimensions
        of the image when on disk so we'll go through a phased approach...
        1. get a list of ALL images from the parent node
        2. filter out any bad image names that we know of (gifs, ads, etc..)
        3. do a head request on each file to make sure it meets our bare requirements
        4. any images left over let's do a full GET request, download em to disk and check their dimensions
        5. Score images based on different factors like height/width and possibly things like color density"""
        good_images = self.get_image_candidates(node)

        if good_images:
            scored_images = self.fetch_images(good_images, parent_depth_level)
            if scored_images:
                highscore_image = sorted(list(scored_images.items()), key=lambda x: x[1], reverse=True)[0][0]
                main_image = Image()
                main_image._src = highscore_image.src
                main_image._width = highscore_image.width
                main_image._height = highscore_image.height
                main_image._extraction_type = "bigimage"
                score_len = len(scored_images)
                main_image._confidence_score = 100 / score_len if score_len > 0 else 0
                return main_image

        depth_obj = self.get_depth_level(node, parent_depth_level, sibling_depth_level)
        if depth_obj:
            return self.check_large_images(depth_obj.node, depth_obj.parent_depth, depth_obj.sibling_depth)

        return None

    def get_depth_level(self, node, parent_depth, sibling_depth):
        max_parent_depth = 2
        if parent_depth > max_parent_depth:
            return None

        sibling_node = self.parser.previous_sibling(node)
        if sibling_node is not None:
            return DepthTraversal(sibling_node, parent_depth, sibling_depth + 1)
        if node is not None:
            parent = self.parser.get_parent(node)
            if parent is not None:
                return DepthTraversal(parent, parent_depth + 1, 0)
        return None

    def fetch_images(self, images, depth_level):
        """download the images to temp disk and set their dimensions
        - we're going to score the images in the order in which they appear so images higher up will have more
          importance,
        - we'll count the area of the 1st image as a score of 1 and then calculate how much larger or small
          each image after it is
        - we'll also make sure to try and weed out banner type ad blocks that have big widths and small heights
          or vice versa
        - so if the image is 3rd found in the dom it's sequence score would be 1 / 3 = .33 * diff in area from
          the first image"""
        image_results = {}
        initial_area = float(0.0)
        total_score = float(0.0)
        cnt = float(1.0)
        min_width = 50
        for image in images[:30]:
            src = self.parser.get_attribute(image, attr="src")
            src = self.build_image_path(src)
            src = self.add_schema_if_none(src)
            local_image = self.get_local_image(src)
            width = local_image.width
            height = local_image.height
            src = local_image.src
            file_extension = local_image.file_extension

            if file_extension != ".gif" or file_extension != "NA":
                if (depth_level >= 1 and local_image.width > 300) or depth_level < 1:
                    if not self.is_banner_dimensions(width, height):
                        if width > min_width:
                            sequence_score = float(1.0 / cnt)
                            area = float(width * height)
                            total_score = float(0.0)

                            if initial_area == 0:
                                initial_area = area * float(1.48)
                                total_score = 1.0
                            else:
                                area_difference = float(area / initial_area)
                                total_score = sequence_score * area_difference

                            image_results.update({local_image: total_score})
                            cnt += 1
        return image_results

    def get_image(self, src, score=100, extraction_type="N/A"):
        # build the Image object
        image = Image()
        image._src = self.build_image_path(src)
        image._extraction_type = extraction_type
        image._confidence_score = score

        # check if we have a local image
        # in order to add more information
        # on the Image object
        local_image = self.get_local_image(image.src)
        if local_image:
            image._bytes = local_image.bytes
            image._height = local_image.height
            image._width = local_image.width

        # return the image
        return image

    @staticmethod
    def is_banner_dimensions(width, height):
        """returns true if we think this is kind of a bannery dimension like 600 / 100 = 6 may be a fishy
        dimension for a good image"""
        if width == height:
            return False

        if width > height:
            diff = float(width / height)
            if diff > 5:
                return True

        if height > width:
            diff = float(height / width)
            if diff > 5:
                return True

        return False

    def get_node_images(self, node):
        images = self.parser.get_elements_by_tag(node, tag="img")
        if images is not None and len(images) < 1:
            return None
        return images

    def filter_bad_names(self, images):
        """takes a list of image elements and filters out the ones with bad names"""
        good_images = []
        for image in images:
            if self.is_valid_filename(image):
                good_images.append(image)
        return good_images if len(good_images) > 0 else None

    def is_valid_filename(self, image_node):
        """will check the image src against a list of bad image files we know of like buttons, etc..."""
        src = self.parser.get_attribute(image_node, attr="src")

        if not src:
            return False

        if self.badimages_names_re.search(src):
            return False

        return True

    def get_image_candidates(self, node):
        good_images = []
        filtered_images = []
        images = self.get_node_images(node)
        if images:
            filtered_images = self.filter_bad_names(images)
        if filtered_images:
            good_images = self.get_images_bytesize_match(filtered_images)
        return good_images

    def get_images_bytesize_match(self, images):
        """loop through all the images and find the ones that have the best bytez to even make them a candidate"""
        cnt = 0
        max_bytes_size = 15728640
        good_images = []
        for image in images:
            if cnt > 30:
                return good_images
            src = self.parser.get_attribute(image, attr="src")
            src = self.build_image_path(src)
            src = self.add_schema_if_none(src)
            local_image = self.get_local_image(src)
            if local_image:
                filesize = local_image.bytes
                if (filesize == 0 or filesize > self.images_min_bytes) and filesize < max_bytes_size:
                    good_images.append(image)
                else:
                    images.remove(image)
            cnt += 1
        return good_images if len(good_images) > 0 else None

    @staticmethod
    def get_node(node):
        return node if node else None

    def check_link_tag(self):
        """checks to see if we were able to find open link_src on this page"""
        node = self.article.raw_doc
        meta = self.parser.get_elements_by_tag(node, tag="link", attr="rel", value="image_src")
        for item in meta:
            src = self.parser.get_attribute(item, attr="href")
            if src:
                return self.get_image(src, extraction_type="linktag")
        return None

    def check_known_schemas(self):
        """checks to see if we were able to find the image via known schemas:

        Supported Schemas
         - Open Graph
         - schema.org"""
        if "image" in self.article.opengraph:
            return self.get_image(self.article.opengraph["image"], extraction_type="opengraph")
        if self.article.schema and "image" in self.article.schema and "url" in self.article.schema["image"]:
            return self.get_image(self.article.schema["image"]["url"], extraction_type="schema.org")
        return None

    def get_local_image(self, src):
        """returns the bytes of the image file on disk"""
        return ImageUtils.store_image(self.fetcher, self.article.link_hash, src, self.config)

    def get_clean_domain(self):
        if self.article.domain:
            return self.article.domain.replace("www.", "")
        return None

    def check_known_elements(self):
        """in here we check for known image contains from sites we've checked out like yahoo,
        techcrunch, etc... that have
        * known  places to look for good images.
        * TODO: enable this to use a series of settings files so people can define what the image ids/classes are on
                specific sites"""
        domain = self.get_clean_domain()
        if domain in self.custom_site_mapping:
            classes = self.custom_site_mapping.get(domain).split("|")
            for classname in classes:
                KNOWN_IMG_DOM_NAMES.append(classname)

        image = None
        doc = self.article.raw_doc

        def _check_elements(elements):
            for element in elements:
                tag = self.parser.get_tag(element)
                if tag == "img":
                    return element
                images = self.parser.get_elements_by_tag(element, tag="img")
                if images:
                    return images[0]
            return None

        # check for elements with known id
        for css in KNOWN_IMG_DOM_NAMES:
            elements = self.parser.get_elements_by_tag(doc, attr="id", value=css)
            image = _check_elements(elements)
            if image is not None:
                src = self.parser.get_attribute(image, attr="src")
                if src:
                    return self.get_image(src, score=90, extraction_type="known")

        # check for elements with known classes
        for css in KNOWN_IMG_DOM_NAMES:
            elements = self.parser.get_elements_by_tag(doc, attr="class", value=css)
            image = _check_elements(elements)
            if image is not None:
                src = self.parser.get_attribute(image, attr="src")
                if src:
                    return self.get_image(src, score=90, extraction_type="known")

        return None

    def build_image_path(self, src):
        """This method will take an image path and build out the absolute path to that image
        * using the initial url we crawled so we can find a link to the image if they use relative urls like
        ../myimage.jpg"""
        o = urlparse(src)
        # we have a full url
        if o.netloc != "":
            return o.geturl()
        # we have a relative url
        return urljoin(self.article.final_url, src)

    def load_customesite_mapping(self):
        path = os.path.join("resources", "images", "known-image-css.txt")
        data_file = FileHelper.load_resource_file(path)
        lines = data_file.splitlines()
        for line in lines:
            domain, css = line.split("^")
            self.custom_site_mapping.update({domain: css})

    def add_schema_if_none(self, src):
        src_test = urlparse(src)
        if src_test.scheme == "":
            target = urlparse(self.article.final_url)
            return str(target.scheme) + ":" + src
        return src



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import base64
import logging
import os

from PIL import Image

from goose3.image import ImageDetails, LocallyStoredImage
from goose3.utils import fnv_1a
from goose3.utils.encoding import smart_str

logger = logging.getLogger(__name__)


class ImageUtils:
    @classmethod
    def get_image_dimensions(cls, path):
        image_details = ImageDetails()
        try:
            # workaround to force the file to actually be closed by Pillow
            with open(path, "rb") as img_file:
                with Image.open(img_file) as image:
                    image_details.set_mime_type(image.format)
                    width, height = image.size
            image_details.set_width(width)
            image_details.set_height(height)
        except OSError:
            image_details.set_mime_type("NA")
            logger.exception("Cannot identify image file")
        return image_details

    @classmethod
    def store_image(cls, http_client, link_hash, src, config):
        """Writes an image src http string to disk as a temporary file and returns the LocallyStoredImage object
        that has the info you should need on the image"""
        # check for a cache hit already on disk
        image = cls.read_localfile(link_hash, src, config)
        if image:
            return image

        # no cache found; do something else

        # parse base64 image
        if src.startswith("data:image"):
            image = cls.write_localfile_base64(link_hash, src, config)
            return image

        # download the image
        data = http_client.fetch(src)
        if data:
            image = cls.write_localfile(data, link_hash, src, config)
            if image:
                return image

        return None

    @classmethod
    def get_mime_type(cls, image_details):
        mime_type = image_details.get_mime_type().lower()
        mimes = {
            "png": ".png",
            "jpg": ".jpg",
            "jpeg": ".jpg",
            "gif": ".gif",
        }
        return mimes.get(mime_type, "NA")

    @classmethod
    def read_localfile(cls, link_hash, src, config):
        local_image_name = cls.get_localfile_name(link_hash, src, config)
        if os.path.isfile(local_image_name):
            image_details = cls.get_image_dimensions(local_image_name)
            file_extension = cls.get_mime_type(image_details)
            filesize = os.path.getsize(local_image_name)
            return LocallyStoredImage(
                src=src,
                local_filename=local_image_name,
                link_hash=link_hash,
                size=filesize,
                file_extension=file_extension,
                height=image_details.get_height(),
                width=image_details.get_width(),
            )
        return None

    @classmethod
    def write_localfile(cls, entity, link_hash, src, config):
        local_path = cls.get_localfile_name(link_hash, src, config)
        with open(local_path, "wb") as fobj:
            fobj.write(entity)
        return cls.read_localfile(link_hash, src, config)

    @classmethod
    def write_localfile_base64(cls, link_hash, src, config):
        data = src[src.find("base64,") + 7 :]
        entity = bytes(base64.b64decode(data))
        return cls.write_localfile(entity, link_hash, src, config)

    @classmethod
    def get_localfile_name(cls, link_hash, src, config):
        image_hash = fnv_1a(smart_str(src))
        return os.path.join(config.local_storage_path, f"{link_hash}_{image_hash}")

    @classmethod
    def clean_src_string(cls, src):
        return src.replace(" ", "%20")

    @classmethod
    def fetch(cls, http_client, src):
        return http_client.fetch(src)



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import codecs
import os
import unittest

from goose3.parsers import Parser, ParserSoup
from goose3.utils.constants import CAMEL_CASE_DEPRICATION
from tests.test_base import fail_after

CURRENT_PATH = os.path.dirname(os.path.abspath(__file__))


def load_resource(path):
    try:
        with codecs.open(path, "r", "utf-8") as fobj:
            content = fobj.read()
        return content
    except OSError:
        raise OSError("Couldn't open file %s" % path)


class ParserBase(unittest.TestCase):
    def setUp(self):
        self.parser = Parser

    def get_html(self, filename):
        path = os.path.join(CURRENT_PATH, "data", filename)
        path = os.path.abspath(path)
        return load_resource(path)

    def test_cssselect(self):
        html = "<html><body>"
        html += '<p class="link">this is a test <a class="link">link</a> and this is <strong class="foo">strong</strong></p>'
        html += '<p>this is a test and this is <strong class="link">strong</strong></p>'
        html += "</body></html>"
        doc = self.parser.fromstring(html)
        # find node with a class attribute
        items_result = self.parser.css_select(doc, "*[class]")
        self.assertEqual(len(items_result), 4)

        # find p nodes
        items_result = self.parser.css_select(doc, "p")
        self.assertEqual(len(items_result), 2)

        # find nodes with attribute class equal to link
        items_result = self.parser.css_select(doc, "*[class=link]")
        self.assertEqual(len(items_result), 3)

        # find p nodes with class attribute
        items_result = self.parser.css_select(doc, "p[class]")
        self.assertEqual(len(items_result), 1)

        # find p nodes with class attribute link
        items_result = self.parser.css_select(doc, "p[class=link]")
        self.assertEqual(len(items_result), 1)

        # find strong nodes with class attribute link or foo
        items_result = self.parser.css_select(doc, "strong[class=link], strong[class=foo]")
        self.assertEqual(len(items_result), 2)

        # find strong nodes with class attribute link or foo
        items_result = self.parser.css_select(doc, "p > a")
        self.assertEqual(len(items_result), 1)

    def test_get_elements_by_tag(self):
        html = "<html><body>"
        html += '<p>this is a test <a class="link">link</a> and this is <strong class="link">strong</strong></p>'
        html += '<p>this is a test and this is <strong class="link">strong</strong></p>'
        html += "</body></html>"
        doc = self.parser.fromstring(html)
        p = self.parser.get_elements_by_tag(doc, tag="p")[0]

        self.assertEqual(len(p), 2)

    def test_replace_tag(self):
        html = self.get_html("parser/test1.html")
        doc = self.parser.fromstring(html)

        # replace all p with div
        ps = self.parser.get_elements_by_tag(doc, tag="p")
        divs = self.parser.get_elements_by_tag(doc, tag="div")
        pcount = len(ps)
        divcount = len(divs)
        for p in ps:
            self.parser.replace_tag(p, "div")
        divs2 = self.parser.get_elements_by_tag(doc, tag="div")
        divcount2 = len(divs2)
        self.assertEqual(divcount2, pcount + divcount)

        # replace first div span with center
        spans = self.parser.get_elements_by_tag(doc, tag="span")
        spanscount = len(spans)
        div = self.parser.get_elements_by_tag(doc, tag="div")[0]
        span = self.parser.get_elements_by_tag(div, tag="span")
        self.assertEqual(len(span), 1)
        self.parser.replace_tag(span[0], "center")
        span = self.parser.get_elements_by_tag(div, tag="span")
        self.assertEqual(len(span), 0)
        centers = self.parser.get_elements_by_tag(div, tag="center")
        self.assertEqual(len(centers), 1)

    def test_droptag(self):
        # test with 1 node
        html = "<html><body><div>Hello <b>World!</b></div></body></html>"
        expecte_html = "<html><body><div>Hello World!</div></body></html>"
        doc = self.parser.fromstring(html)
        nodes = self.parser.css_select(doc, "b")
        self.assertEqual(len(nodes), 1)
        self.parser.drop_tag(nodes)

        nodes = self.parser.css_select(doc, "b")
        self.assertEqual(len(nodes), 0)

        result_html = self.parser.node_to_string(doc)
        self.assertEqual(expecte_html, result_html)

        # test with 2 nodes
        html = "<html><body><div>Hello <b>World!</b> bla <b>World!</b></div></body></html>"
        expecte_html = "<html><body><div>Hello World! bla World!</div></body></html>"
        doc = self.parser.fromstring(html)
        nodes = self.parser.css_select(doc, "b")
        self.assertEqual(len(nodes), 2)
        self.parser.drop_tag(nodes)

        nodes = self.parser.css_select(doc, "b")
        self.assertEqual(len(nodes), 0)

        result_html = self.parser.node_to_string(doc)
        self.assertEqual(expecte_html, result_html)

    def test_tostring(self):
        html = "<html><body>"
        html += "<p>this is a test <a>link</a> and this is <strong>strong</strong></p>"
        html += "</body></html>"
        doc = self.parser.fromstring(html)
        result = self.parser.node_to_string(doc)
        self.assertEqual(html, result)

    def test_strip_tags(self):
        html = "<html><body>"
        html += "<p>this is a test <a>link</a> and this is <strong>strong</strong></p>"
        html += "</body></html>"
        expected = "<html><body>"
        expected += "<p>this is a test link and this is strong</p>"
        expected += "</body></html>"
        doc = self.parser.fromstring(html)
        self.parser.strip_tags(doc, "a", "strong")
        result = self.parser.node_to_string(doc)
        self.assertEqual(expected, result)

    def test_get_elements_by_tags(self):
        html = "<html><body>"
        html += '<p>this is a test <a class="link">link</a> and this is <strong class="link">strong</strong></p>'
        html += '<p>this is a test and this is <strong class="link">strong</strong></p>'
        html += "</body></html>"
        doc = self.parser.fromstring(html)
        elements = self.parser.get_elements_by_tags(doc, ["p", "a", "strong"])
        self.assertEqual(len(elements), 5)

        # find childs within the first p
        p = self.parser.get_elements_by_tag(doc, tag="p")[0]
        elements = self.parser.get_elements_by_tags(p, ["p", "a", "strong"])
        self.assertEqual(len(elements), 2)

    def test_get_elements_by_tag(self):
        html = "<html><body>"
        html += "<p>this is a test <a>link</a> and this is <strong>strong</strong></p>"
        html += "</body></html>"
        doc = self.parser.fromstring(html)
        # find all tags
        elements = self.parser.get_elements_by_tag(doc)
        self.assertEqual(len(elements), 5)

        # find all p
        elements = self.parser.get_elements_by_tag(doc, tag="p")
        self.assertEqual(len(elements), 1)

        html = "<html><body>"
        html += '<p>this is a test <a class="link classB classc">link</a> and this is <strong class="link">strong</strong></p>'
        html += '<p>this is a test and this is <strong class="Link">strong</strong></p>'
        html += "</body></html>"
        doc = self.parser.fromstring(html)
        # find all p
        elements = self.parser.get_elements_by_tag(doc, tag="p")
        self.assertEqual(len(elements), 2)

        # find all a
        elements = self.parser.get_elements_by_tag(doc, tag="a")
        self.assertEqual(len(elements), 1)

        # find all strong
        elements = self.parser.get_elements_by_tag(doc, tag="strong")
        self.assertEqual(len(elements), 2)

        # find first p
        # and find strong elemens within the p
        elem = self.parser.get_elements_by_tag(doc, tag="p")[0]
        elements = self.parser.get_elements_by_tag(elem, tag="strong")
        self.assertEqual(len(elements), 1)

        # test if the first p in taken in account
        elem = self.parser.get_elements_by_tag(doc, tag="p")[0]
        elements = self.parser.get_elements_by_tag(elem, tag="p")
        self.assertEqual(len(elements), 0)

        # find elem with class "link"
        elements = self.parser.get_elements_by_tag(doc, attr="class", value="link")
        self.assertEqual(len(elements), 3)

        # find elem with class "classB"
        elements = self.parser.get_elements_by_tag(doc, attr="class", value="classB")
        self.assertEqual(len(elements), 1)

        # find elem with class "classB"
        elements = self.parser.get_elements_by_tag(doc, attr="class", value="classc")
        self.assertEqual(len(elements), 1)

        # find elem with class "link" with tag strong
        elements = self.parser.get_elements_by_tag(doc, tag="strong", attr="class", value="link")
        self.assertEqual(len(elements), 2)

        # find elem with class "link" with tag strong
        # within the second p
        elem = self.parser.get_elements_by_tag(doc, tag="p")[1]
        elements = self.parser.get_elements_by_tag(elem, tag="strong", attr="class", value="link")
        self.assertEqual(len(elements), 1)

    def test_del_attribute(self):
        html = self.get_html("parser/test1.html")
        doc = self.parser.fromstring(html)

        # find div element with class foo
        elements = self.parser.get_elements_by_tag(doc, tag="div", attr="class", value="foo")
        self.assertEqual(len(elements), 1)

        div = elements[0]
        # remove an unexistant attribute
        self.parser.del_attribute(div, attr="bla")
        # remove the attribute class
        self.parser.del_attribute(div, attr="class")

        # find div element with class foo
        elements = self.parser.get_elements_by_tag(doc, tag="div", attr="class", value="foo")
        self.assertEqual(len(elements), 0)

    def test_encoding(self):
        """
        If pass unicode string to lxml.html.fromstring with encoding set in document will receive:
        "ValueError: Unicode strings with encoding declaration are not supported.
        Please use bytes input or XML fragments without declaration."
        Test for this case.
        """
        html = """
        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
        """
        html += "<html><body>"
        html += "<p>Я рядочок</p>"
        html += "</body></html>"
        self.parser.fromstring(html)

    @fail_after(CAMEL_CASE_DEPRICATION, "nodeToString")
    def test_tostring_old(self):
        html = "<html><body>"
        html += "<p>this is a test <a>link</a> and this is <strong>strong</strong></p>"
        html += "</body></html>"
        doc = self.parser.fromstring(html)
        result = self.parser.nodeToString(doc)
        self.assertEqual(html, result)

    @fail_after(CAMEL_CASE_DEPRICATION, "replaceTag")
    def test_replace_tag_old(self):
        html = self.get_html("parser/test1.html")
        doc = self.parser.fromstring(html)

        # replace all p with div
        ps = self.parser.get_elements_by_tag(doc, tag="p")
        divs = self.parser.get_elements_by_tag(doc, tag="div")
        pcount = len(ps)
        divcount = len(divs)
        for p in ps:
            self.parser.replaceTag(p, "div")
        divs2 = self.parser.get_elements_by_tag(doc, tag="div")
        divcount2 = len(divs2)
        self.assertEqual(divcount2, pcount + divcount)

    @fail_after(CAMEL_CASE_DEPRICATION, "stripTags")
    def test_strip_tags_old(self):
        html = "<html><body>"
        html += "<p>this is a test <a>link</a> and this is <strong>strong</strong></p>"
        html += "</body></html>"
        expected = "<html><body>"
        expected += "<p>this is a test link and this is strong</p>"
        expected += "</body></html>"
        doc = self.parser.fromstring(html)
        self.parser.stripTags(doc, "a", "strong")
        result = self.parser.node_to_string(doc)
        self.assertEqual(expected, result)

    @fail_after(CAMEL_CASE_DEPRICATION, "getElementById")
    def test_get_element_by_id_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "getElementsByTag")
    def test_get_elements_by_tag_old(self):
        html = "<html><body>"
        html += '<p>this is a test <a class="link">link</a> and this is <strong class="link">strong</strong></p>'
        html += '<p>this is a test and this is <strong class="link">strong</strong></p>'
        html += "</body></html>"
        doc = self.parser.fromstring(html)
        p = self.parser.getElementsByTag(doc, tag="p")[0]

        self.assertEqual(len(p), 2)

    @fail_after(CAMEL_CASE_DEPRICATION, "appendChild")
    def test_append_child_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "childNodes")
    def test_child_nodes_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "childNodesWithText")
    def test_child_nodes_with_text_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "textToPara")
    def test_text_to_para_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "getChildren")
    def test_get_children_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "getElementsByTags")
    def test_get_elements_by_tags_old(self):
        html = "<html><body>"
        html += '<p>this is a test <a class="link">link</a> and this is <strong class="link">strong</strong></p>'
        html += '<p>this is a test and this is <strong class="link">strong</strong></p>'
        html += "</body></html>"
        doc = self.parser.fromstring(html)
        elements = self.parser.getElementsByTags(doc, ["p", "a", "strong"])
        self.assertEqual(len(elements), 5)

        # find childs within the first p
        p = self.parser.get_elements_by_tag(doc, tag="p")[0]
        elements = self.parser.getElementsByTags(p, ["p", "a", "strong"])
        self.assertEqual(len(elements), 2)

    @fail_after(CAMEL_CASE_DEPRICATION, "createElement")
    def test_create_element_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "getComments")
    def test_get_comments_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "getParent")
    def test_get_parent_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "getTag")
    def test_get_tag_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "getText")
    def test_get_text_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "previousSiblings")
    def test_previous_siblings_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "nextSibling")
    def test_next_sibling_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "isTextNode")
    def test_is_text_node_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "getAttribute")
    def test_get_attribute_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "delAttribute")
    def test_del_attribute_old(self):
        html = self.get_html("parser/test1.html")
        doc = self.parser.fromstring(html)

        # find div element with class foo
        elements = self.parser.get_elements_by_tag(doc, tag="div", attr="class", value="foo")
        self.assertEqual(len(elements), 1)

        div = elements[0]
        # remove an unexistant attribute
        self.parser.del_attribute(div, attr="bla")
        # remove the attribute class
        self.parser.del_attribute(div, attr="class")

        # find div element with class foo
        elements = self.parser.get_elements_by_tag(doc, tag="div", attr="class", value="foo")
        self.assertEqual(len(elements), 0)

    @fail_after(CAMEL_CASE_DEPRICATION, "setAttribute")
    def test_set_attribute_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)

    @fail_after(CAMEL_CASE_DEPRICATION, "outerHtml")
    def test_outer_html_old(self):  # TODO: currently there isn't one that tests this functionality directly
        self.assertTrue(True)


class TestParser(ParserBase):
    pass


class TestSoupParser(ParserBase):
    def setUp(self):
        self.parser = ParserSoup



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import unittest

from goose3.text import (
    encode_value,
    encodeValue,
    get_encodings_from_content,
    inner_trim,
    innerTrim,
)
from goose3.utils.constants import CAMEL_CASE_DEPRICATION

from .test_base import fail_after


class TestText(unittest.TestCase):
    def test_get_encodings_from_content_with_trail_spaces(self):
        self.assertEqual(get_encodings_from_content('<meta charset=utf-8 " />'), ["utf-8"])
        self.assertEqual(
            get_encodings_from_content('<meta http-equiv="content-type" content="text/html; charset=utf-8 " />'),
            ["utf-8"],
        )
        self.assertEqual(
            get_encodings_from_content('<meta http-equiv="content-type" content="text/html; charset=utf-8    " />'),
            ["utf-8"],
        )
        self.assertEqual(
            get_encodings_from_content(b'<meta http-equiv="content-type" content="text/html; charset=utf-8 " />'),
            ["utf-8"],
        )

    def test_get_encodings_from_content_with_out_trail_spaces(self):
        self.assertEqual(
            get_encodings_from_content('<meta http-equiv="content-type" content="text/html; charset=utf-8" />'),
            ["utf-8"],
        )
        self.assertEqual(
            get_encodings_from_content(b'<meta http-equiv="content-type" content="text/html; charset=utf-8" />'),
            ["utf-8"],
        )

    def test_inner_trim(self):
        self.assertEqual(inner_trim("Some    Crazy  looking text"), "Some Crazy looking text")

    def test_encode_value(self):
        self.assertEqual(encode_value(b"something here"), "something here")

    @fail_after(CAMEL_CASE_DEPRICATION, "innerTrim")
    def test_inner_trim_old(self):
        self.assertEqual(innerTrim("Some    Crazy  looking text"), "Some Crazy looking text")

    @fail_after(CAMEL_CASE_DEPRICATION, "encodeValue")
    def test_encode_value_old(self):
        self.assertEqual(encodeValue(b"something here"), "something here")



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from .test_base import TestExtractionBase


class TestArticleTags(TestExtractionBase):
    def assert_tags(self, field, expected_value, result_value):
        """ """
        # as we have a set in expected_value and a list in result_value
        # make result_value a set
        expected_value = set(expected_value)

        # check if both have the same number of items
        msg = "expected tags set and result tags set" "don't have the same number of items"
        self.assertEqual(len(result_value), len(expected_value), msg=msg)

        # check if each tag in result_value is in expected_value
        for tag in result_value:
            self.assertTrue(tag in expected_value)

    def test_tags_kexp(self):
        article = self.getArticle()
        fields = ["tags"]
        self.runArticleAssertions(article=article, fields=fields)

    def test_tags_deadline(self):
        article = self.getArticle()
        fields = ["tags"]
        self.runArticleAssertions(article=article, fields=fields)

    def test_tags_wnyc(self):
        article = self.getArticle()
        fields = ["tags"]
        self.runArticleAssertions(article=article, fields=fields)

    def test_tags_cnet(self):
        article = self.getArticle()
        fields = ["tags"]
        self.runArticleAssertions(article=article, fields=fields)

    def test_tags_abcau(self):
        """
        Test ABC Australia page with "topics" tags
        """
        article = self.getArticle()
        fields = ["tags"]
        self.runArticleAssertions(article=article, fields=fields)



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import codecs
import functools
import json
import os
import unittest
from datetime import datetime

import requests_mock

from goose3 import Goose
from goose3.configuration import Configuration
from goose3.version import __version__

CURRENT_PATH = os.path.dirname(os.path.abspath(__file__))


def load_resource(path):
    try:
        with codecs.open(path, "r", "utf-8") as fobj:
            content = fobj.read()
        return content
    except OSError:
        raise OSError("Couldn't open file %s" % path)


def fail_after(version, method=None):
    """Decorator to add to tests to ensure that they fail if a deprecated
    feature is not removed before the specified version
    Args:
        version (str): The version to check against"""

    def decorator_wrapper(func):
        @functools.wraps(func)
        def test_inner(*args, **kwargs):
            if [int(x) for x in version.split(".")] <= [int(x) for x in __version__.split(".")]:
                msg = "{} is depricated and must be removed by version {}".format(
                    method if method else func.__name__, version
                )
                raise AssertionError(msg)
            return func(*args, **kwargs)

        return test_inner

    return decorator_wrapper


class TestExtractionBase(unittest.TestCase):
    """
    Extraction test case
    """

    # callback = MockResponseExtractors

    def getRawHtml(self):
        test, suite, module, cls, func = self.id().split(".")
        path = os.path.join(
            os.path.dirname(CURRENT_PATH), "data", suite, module.partition("test_")[2], "%s.html" % func
        )
        path = os.path.abspath(path)
        content = load_resource(path)
        return content

    def loadData(self):
        """ """
        full_id = self.id().split(".")

        test, module, cls, func = full_id
        path = os.path.join(
            os.path.dirname(CURRENT_PATH),
            "tests",
            "data",
            # suite,
            module.partition("test_")[2],
            "%s.json" % func,
        )
        path = os.path.abspath(path)
        content = load_resource(path)
        self.data = json.loads(content)

    def loadHtml(self):
        test, module, cls, func = self.id().split(".")
        path = os.path.join(
            os.path.dirname(CURRENT_PATH), "tests", "data", module.partition("test_")[2], "%s.html" % func
        )
        path = os.path.abspath(path)
        self.html = load_resource(path)

    def assert_cleaned_text(self, field, expected_value, result_value):
        # # TODO : handle verbose level in tests
        # print "\n=======================::. ARTICLE REPORT %s .::======================\n" % self.id()
        # print 'expected_value (%s) \n' % len(expected_value)
        # print expected_value
        # print "-------"
        # print 'result_value (%s) \n' % len(result_value)
        # print result_value

        # cleaned_text is Null
        msg = "Resulting article text was NULL!"
        self.assertNotEqual(result_value, None, msg=msg)

        # cleaned_text length
        msg = "Article text was not as long as expected beginning!"
        self.assertTrue(len(expected_value) <= len(result_value), msg=msg)

        # clean_text value
        result_value = result_value[0 : len(expected_value)]
        msg = "The beginning of the article text was not as expected!"
        self.assertEqual(expected_value, result_value, msg=msg)

    def runArticleAssertions(self, article, fields):
        """ """
        for field in fields:
            expected_value = self.data["expected"][field]
            result_value = getattr(article, field, None)

            # handle checking datetimes...
            if field in ["publish_datetime_utc"]:
                self.assertEqual(type(result_value), type(datetime.today()))
                result_value = result_value.isoformat(sep=" ")

            # custom assertion for a given field
            assertion = "assert_%s" % field
            if hasattr(self, assertion):
                getattr(self, assertion)(field, expected_value, result_value)
                continue

            # default assertion
            msg = f"Error {field} \nexpected: {expected_value}\nresult:   {result_value}"
            self.assertEqual(expected_value, result_value, msg=msg)

    # def extract(self, instance):
    #     article_url = self.data['url']
    #     with requests_mock.mock() as m:
    #         for url, content in self.callback(self).contents():
    #             m.get(url, content=content)
    #         article = instance.extract(url=article_url)
    #         return article

    def getConfig(self):
        config = Configuration()
        config.enable_image_fetching = False
        return config

    def getArticle(self, config_=None):
        """ """
        # load test case data
        self.loadData()
        self.loadHtml()

        # basic configuration
        # no image fetching
        config = self.getConfig()
        if config is not None:
            if isinstance(config_, dict):
                for k, v in list(config_.items()):
                    if hasattr(config, k):
                        setattr(config, k, v)
        self.parser = config.get_parser()

        # target language
        # needed for non english language most of the time
        target_language = self.data.get("target_language")
        if target_language:
            config.target_language = target_language
            config.use_meta_language = False

        # read in the basic image...
        with open(f"{CURRENT_PATH}/data/images/50850547cc7310bc53e30e802c6318f1", "rb") as fobj:
            img_content = fobj.read()

        # read in another, blank image
        with open(f"{CURRENT_PATH}/data/images/blank.jpeg", "rb") as fobj:
            blank_img = fobj.read()

        # run goose
        with Goose(config=config) as g, requests_mock.Mocker(real_http=False) as m:
            # load images for those tests
            m.get("http://go.com/images/465395/", content=blank_img)
            m.get("http://bla.com/images/465395/", content=blank_img)
            m.get(
                "http://md0.libe.com/photo/465395/?modified_at=1351411813&ratio_x=03&ratio_y=02&width=476",
                content=img_content,
            )
            # if the url is not given in the result json, use the raw_html parameter.
            if "url" in self.data:
                m.get(self.data["url"], text=self.html)
                return g.extract(url=self.data["url"])
            else:
                return g.extract(raw_html=self.html)



"""Handle Encoding"""
import datetime
from decimal import Decimal


class DjangoUnicodeDecodeError(UnicodeDecodeError):
    def __init__(self, obj, *args):
        self.obj = obj
        UnicodeDecodeError.__init__(self, *args)

    def __str__(self):
        original = UnicodeDecodeError.__str__(self)
        return f"{original}. You passed in {self.obj!r} ({type(self.obj)})"


def smart_unicode(string, encoding="utf-8", strings_only=False, errors="strict"):
    """Returns a unicode object representing 's'. Treats bytestrings using the 'encoding' codec.

    If strings_only is True, don't convert (some) non-string-like objects.
    """
    # if isinstance(s, Promise):
    #     # The input is the result of a get_text_lazy() call.
    #     return s
    return force_unicode(string, encoding, strings_only, errors)


def is_protected_type(obj):
    """Determine if the object instance is of a protected type.

    Objects of protected types are preserved as-is when passed to force_unicode(strings_only=True).
    """
    return isinstance(obj, (type(None), int, datetime.datetime, datetime.date, datetime.time, float, Decimal))


def force_unicode(string, encoding="utf-8", strings_only=False, errors="strict"):
    """Similar to smart_unicode, except that lazy instances are resolved to strings, rather than kept as lazy objects.

    If strings_only is True, don't convert (some) non-string-like objects.
    """
    # Handle the common case first, saves 30-40% in performance when s
    # is an instance of unicode. This function gets called often in that
    # setting.
    if isinstance(string, str):
        return string
    if strings_only and is_protected_type(string):
        return string
    try:
        if not isinstance(string, str):
            if hasattr(string, "__unicode__"):
                string = string.__unicode__()
            else:
                try:
                    string = str(string, encoding, errors)
                except UnicodeEncodeError:
                    if not isinstance(string, Exception):
                        raise
                    # If we get to here, the caller has passed in an Exception
                    # subclass populated with non-ASCII data without special
                    # handling to display as a string. We need to handle this
                    # without raising a further exception. We do an
                    # approximation to what the Exception's standard str()
                    # output should be.
                    string = " ".join([force_unicode(arg, encoding, strings_only, errors) for arg in string])
        elif not isinstance(string, str):
            # Note: We use .decode() here, instead of unicode(s, encoding,
            # errors), so that if s is a SafeString, it ends up being a
            # SafeUnicode at the end.
            string = string.decode(encoding, errors)
    except UnicodeDecodeError as ex:
        if not isinstance(string, Exception):
            raise DjangoUnicodeDecodeError(string, *ex.args) from ex
        # If we get to here, the caller has passed in an Exception
        # subclass populated with non-ASCII bytestring data without a
        # working unicode method. Try to handle this without raising a
        # further exception by individually forcing the exception args
        # to unicode.
        string = " ".join([force_unicode(arg, encoding, strings_only, errors) for arg in string])
    return string


def smart_str(string, encoding="utf-8", strings_only=False, errors="strict"):
    """Returns a bytestring version of 's', encoded as specified in 'encoding'.

    If strings_only is True, don't convert (some) non-string-like objects.
    """
    val = string
    if strings_only and isinstance(string, (type(None), int)):
        val = string
    # if isinstance(s, Promise):
    #     return unicode(s).encode(encoding, errors)
    if isinstance(string, str):
        try:
            val = string.encode(encoding, errors)
        except UnicodeEncodeError:
            val = string.encode("utf-8", errors)
    elif not isinstance(string, bytes):
        try:
            val = str(string).encode(encoding, errors)
        except UnicodeEncodeError:
            if isinstance(string, Exception):
                # An Exception subclass containing non-ASCII data that doesn't
                # know how to print itself properly. We shouldn't raise a
                # further exception.
                val = " ".join([smart_str(arg, encoding, strings_only, errors) for arg in string])
            val = str(string).encode(encoding, errors)
    return val



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import html

from goose3.text import inner_trim


class OutputFormatter:
    def __init__(self, config, article):
        # config
        self.config = config

        # article
        self.article = article

        # parser
        self.parser = self.config.get_parser()

        # stopwords class
        self.stopwords_class = config.stopwords_class

        # top node
        self.top_node = None

    def get_language(self):
        """
        Returns the language is by the article or
        the configuration language
        """
        # we don't want to force the target language
        # so we use the article.meta_lang
        if self.config.use_meta_language:
            if self.article.meta_lang:
                return self.article.meta_lang[:2]
        return self.config.target_language

    def get_top_node(self):
        return self.top_node

    def get_formatted_text(self):
        self.top_node = self.article.top_node
        self.remove_negativescores_nodes()
        self.links_to_text()
        self.add_newline_to_br()
        self.replace_with_text()
        self.remove_fewwords_paragraphs()
        self.make_list_elms_pretty()
        return self.convert_to_text()

    def convert_to_text(self):
        txts = []
        for node in list(self.get_top_node()):
            txt = self.parser.get_text(node)
            if txt:
                txt = html.unescape(txt)
                txt_lis = inner_trim(txt).split(r"\n")
                txts.extend(txt_lis)
        text = "\n\n".join(txts)
        # ensure no double newlines at the beginning of lists
        if self.config.parse_lists:
            # Split out the lists and clean them up! Ensuring no trailing spaces
            txt = text.replace("\n•", "•").split("• ")
            txt = [x.strip() for x in txt]

            if self.config.pretty_lists:
                text = "\n• ".join(txt)
            else:
                text = "\n".join(txt)
        return text

    def add_newline_to_br(self):
        for elm in self.parser.get_elements_by_tag(self.top_node, tag="br"):
            elm.text = r"\n"

    def links_to_text(self):
        """cleans up and converts any nodes that should be considered text into text"""
        self.parser.strip_tags(self.get_top_node(), "a")

    def make_list_elms_pretty(self):
        """make any list element read like a list"""
        for elm in self.parser.get_elements_by_tag(self.top_node, tag="li"):
            elm.text = rf"• {elm.text}"

    def remove_negativescores_nodes(self):
        """if there are elements inside our top node that have a negative gravity score, let's give em the boot"""
        gravity_items = self.parser.css_select(self.top_node, "*[gravityScore]")
        for item in gravity_items:
            score = self.parser.get_attribute(item, "gravityScore")
            score = int(score, 0)
            if score < 1:
                item.getparent().remove(item)

    def replace_with_text(self):
        """replace common tags with just text so we don't have any crazy formatting issues
        so replace <br>, <i>, <strong>, etc.... with whatever text is inside them
        code : http://lxml.de/api/lxml.etree-module.html#strip_tags
        """
        self.parser.strip_tags(self.get_top_node(), "b", "strong", "i", "br")
        if self.config.keep_footnotes:
            self.parser.strip_tags(self.get_top_node(), "sup")

    def remove_fewwords_paragraphs(self):
        """remove paragraphs that have less than x number of words, would indicate that it's some sort of link"""
        all_nodes = self.parser.get_elements_by_tags(self.get_top_node(), ["*"])
        all_nodes.reverse()
        for elm in all_nodes:
            tag = self.parser.get_tag(elm)
            text = self.parser.get_text(elm)
            stop_words = self.stopwords_class(language=self.get_language()).get_stopword_count(text)
            if (
                (tag != "br" or text != "\\r")
                and stop_words.get_stopword_count() < 3
                and len(self.parser.get_elements_by_tag(elm, tag="object")) == 0
                and len(self.parser.get_elements_by_tag(elm, tag="embed")) == 0
            ):
                self.parser.remove(elm)
            # TODO
            # check if it is in the right place
            else:
                trimmed = self.parser.get_text(elm)
                if trimmed.startswith("(") and trimmed.endswith(")"):
                    self.parser.remove(elm)


class StandardOutputFormatter(OutputFormatter):
    pass



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
import os
import weakref
from tempfile import mkstemp
from typing import List, Union

from goose3.article import Article  # noqa: F401
from goose3.configuration import (  # noqa: F401
    ArticleContextPattern,
    Configuration,
    PublishDatePattern,
)
from goose3.crawler import CrawlCandidate, Crawler
from goose3.image import Image  # noqa: F401
from goose3.network import NetworkFetcher
from goose3.video import Video  # noqa: F401

logger = logging.getLogger(__name__)


class Goose:
    """Extract most likely article content and aditional metadata from a URL or previously fetched HTML document

    Args:
        config (Configuration, dict): A configuration file or dictionary representation of the configuration file
    Returns:
        Goose: An instance of the goose extraction object"""

    def __init__(self, config: Union[Configuration, dict] = None):
        # Use the passed in configuration if it is of the right type, otherwise
        # use the default as a base
        if isinstance(config, Configuration):
            self.config = config
        else:
            self.config = Configuration()

        # if config was a passed in dict, parse it into the stored configuration
        if isinstance(config, dict):
            for k, v in config.items():
                if hasattr(self.config, k):
                    setattr(self.config, k, v)

        # setup a single network connection
        self.fetcher = NetworkFetcher(self.config)
        self.finalizer = weakref.finalize(self, self.close)

        # we don't need to go further if image extractor or local_storage is not set
        if not self.config.local_storage_path or not self.config.enable_image_fetching:
            return

        # test if config.local_storage_path is a directory
        if not os.path.isdir(self.config.local_storage_path):
            os.makedirs(self.config.local_storage_path)

        if not os.path.isdir(self.config.local_storage_path):
            msg = (
                f"{self.config.local_storage_path} directory does not seem to exist, "
                "you need to set this for image processing downloads"
            )
            raise Exception(msg)

        # test to write a dummy file to the directory to check is directory is writable
        level, path = mkstemp(dir=self.config.local_storage_path)
        try:
            with os.fdopen(level, "w"):
                pass
            os.remove(path)
        except OSError as exc:
            msg = (
                f"{self.config.local_storage_path} directory is not writeble, "
                "you need to set this for image processing downloads"
            )
            raise Exception(msg) from exc

    def __enter__(self):
        """Setup the context manager"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Define what to do when the context manager exits"""
        self.close()

    def close(self):
        """Close the network connection and perform any other required cleanup

        Note:
            Auto closed when using goose as a context manager or when garbage collected"""
        if self.fetcher is not None:
            self.shutdown_network()
        self.finalizer.atexit = False  # turn off the garbage collection close

    def extract(self, url: str = None, raw_html: str = None) -> Article:
        """Extract the most likely article content from the html page

        Args:
            url (str): URL to pull and parse
            raw_html (str): String representation of the HTML page
        Returns:
            Article: Representation of the article contents including other parsed and extracted metadata"""
        if not url and not raw_html:
            raise ValueError("Either url or raw_html should be provided")

        crawl_candidate = CrawlCandidate(self.config, url, raw_html)
        return self.__crawl(crawl_candidate)

    def shutdown_network(self):
        """Close the network connection

        Note:
            Auto closed when using goose as a context manager or when garbage collected"""
        self.fetcher.close()
        self.fetcher = None

    def __crawl(self, crawl_candidate: CrawlCandidate):
        """wrap the crawling functionality"""

        def crawler_wrapper(parser: str, parsers: List[str], crawl_candidate: CrawlCandidate):
            try:
                crawler = Crawler(self.config, self.fetcher)
                article = crawler.crawl(crawl_candidate)
            except (UnicodeDecodeError, ValueError) as ex:
                logger.error("Parser %s failed to parse the content", parser)
                if parsers:
                    parser = parsers.pop(0)  # remove it also!
                    return crawler_wrapper(parser, parsers, crawl_candidate)
                raise ex
            return article

        # use the wrapper
        parsers = list(self.config.available_parsers)
        parsers.remove(self.config.parser_class)
        return crawler_wrapper(self.config.parser_class, parsers, crawl_candidate)



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
from datetime import datetime, timezone

from goose3.extractors import BaseExtractor

# https://stackoverflow.com/a/54629675
TIMEZONE_INFO = {
    "A": 1 * 3600,
    "ACDT": 10.5 * 3600,
    "ACST": 9.5 * 3600,
    "ACT": -5 * 3600,
    "ACWST": 8.75 * 3600,
    "ADT": 4 * 3600,
    "AEDT": 11 * 3600,
    "AEST": 10 * 3600,
    "AET": 10 * 3600,
    "AFT": 4.5 * 3600,
    "AKDT": -8 * 3600,
    "AKST": -9 * 3600,
    "ALMT": 6 * 3600,
    "AMST": -3 * 3600,
    "AMT": -4 * 3600,
    "ANAST": 12 * 3600,
    "ANAT": 12 * 3600,
    "AQTT": 5 * 3600,
    "ART": -3 * 3600,
    "AST": 3 * 3600,
    "AT": -4 * 3600,
    "AWDT": 9 * 3600,
    "AWST": 8 * 3600,
    "AZOST": 0 * 3600,
    "AZOT": -1 * 3600,
    "AZST": 5 * 3600,
    "AZT": 4 * 3600,
    "AoE": -12 * 3600,
    "B": 2 * 3600,
    "BNT": 8 * 3600,
    "BOT": -4 * 3600,
    "BRST": -2 * 3600,
    "BRT": -3 * 3600,
    "BST": 6 * 3600,
    "BTT": 6 * 3600,
    "C": 3 * 3600,
    "CAST": 8 * 3600,
    "CAT": 2 * 3600,
    "CCT": 6.5 * 3600,
    "CDT": -5 * 3600,
    "CEST": 2 * 3600,
    "CET": 1 * 3600,
    "CHADT": 13.75 * 3600,
    "CHAST": 12.75 * 3600,
    "CHOST": 9 * 3600,
    "CHOT": 8 * 3600,
    "CHUT": 10 * 3600,
    "CIDST": -4 * 3600,
    "CIST": -5 * 3600,
    "CKT": -10 * 3600,
    "CLST": -3 * 3600,
    "CLT": -4 * 3600,
    "COT": -5 * 3600,
    "CST": -6 * 3600,
    "CT": -6 * 3600,
    "CVT": -1 * 3600,
    "CXT": 7 * 3600,
    "ChST": 10 * 3600,
    "D": 4 * 3600,
    "DAVT": 7 * 3600,
    "DDUT": 10 * 3600,
    "E": 5 * 3600,
    "EASST": -5 * 3600,
    "EAST": -6 * 3600,
    "EAT": 3 * 3600,
    "ECT": -5 * 3600,
    "EDT": -4 * 3600,
    "EEST": 3 * 3600,
    "EET": 2 * 3600,
    "EGST": 0 * 3600,
    "EGT": -1 * 3600,
    "EST": -5 * 3600,
    "ET": -5 * 3600,
    "F": 6 * 3600,
    "FET": 3 * 3600,
    "FJST": 13 * 3600,
    "FJT": 12 * 3600,
    "FKST": -3 * 3600,
    "FKT": -4 * 3600,
    "FNT": -2 * 3600,
    "G": 7 * 3600,
    "GALT": -6 * 3600,
    "GAMT": -9 * 3600,
    "GET": 4 * 3600,
    "GFT": -3 * 3600,
    "GILT": 12 * 3600,
    "GMT": 0 * 3600,
    "GST": 4 * 3600,
    "GYT": -4 * 3600,
    "H": 8 * 3600,
    "HDT": -9 * 3600,
    "HKT": 8 * 3600,
    "HOVST": 8 * 3600,
    "HOVT": 7 * 3600,
    "HST": -10 * 3600,
    "I": 9 * 3600,
    "ICT": 7 * 3600,
    "IDT": 3 * 3600,
    "IOT": 6 * 3600,
    "IRDT": 4.5 * 3600,
    "IRKST": 9 * 3600,
    "IRKT": 8 * 3600,
    "IRST": 3.5 * 3600,
    "IST": 5.5 * 3600,
    "JST": 9 * 3600,
    "K": 10 * 3600,
    "KGT": 6 * 3600,
    "KOST": 11 * 3600,
    "KRAST": 8 * 3600,
    "KRAT": 7 * 3600,
    "KST": 9 * 3600,
    "KUYT": 4 * 3600,
    "L": 11 * 3600,
    "LHDT": 11 * 3600,
    "LHST": 10.5 * 3600,
    "LINT": 14 * 3600,
    "M": 12 * 3600,
    "MAGST": 12 * 3600,
    "MAGT": 11 * 3600,
    "MART": 9.5 * 3600,
    "MAWT": 5 * 3600,
    "MDT": -6 * 3600,
    "MHT": 12 * 3600,
    "MMT": 6.5 * 3600,
    "MSD": 4 * 3600,
    "MSK": 3 * 3600,
    "MST": -7 * 3600,
    "MT": -7 * 3600,
    "MUT": 4 * 3600,
    "MVT": 5 * 3600,
    "MYT": 8 * 3600,
    "N": -1 * 3600,
    "NCT": 11 * 3600,
    "NDT": 2.5 * 3600,
    "NFT": 11 * 3600,
    "NOVST": 7 * 3600,
    "NOVT": 7 * 3600,
    "NPT": 5.5 * 3600,
    "NRT": 12 * 3600,
    "NST": 3.5 * 3600,
    "NUT": -11 * 3600,
    "NZDT": 13 * 3600,
    "NZST": 12 * 3600,
    "O": -2 * 3600,
    "OMSST": 7 * 3600,
    "OMST": 6 * 3600,
    "ORAT": 5 * 3600,
    "P": -3 * 3600,
    "PDT": -7 * 3600,
    "PET": -5 * 3600,
    "PETST": 12 * 3600,
    "PETT": 12 * 3600,
    "PGT": 10 * 3600,
    "PHOT": 13 * 3600,
    "PHT": 8 * 3600,
    "PKT": 5 * 3600,
    "PMDT": -2 * 3600,
    "PMST": -3 * 3600,
    "PONT": 11 * 3600,
    "PST": -8 * 3600,
    "PT": -8 * 3600,
    "PWT": 9 * 3600,
    "PYST": -3 * 3600,
    "PYT": -4 * 3600,
    "Q": -4 * 3600,
    "QYZT": 6 * 3600,
    "R": -5 * 3600,
    "RET": 4 * 3600,
    "ROTT": -3 * 3600,
    "S": -6 * 3600,
    "SAKT": 11 * 3600,
    "SAMT": 4 * 3600,
    "SAST": 2 * 3600,
    "SBT": 11 * 3600,
    "SCT": 4 * 3600,
    "SGT": 8 * 3600,
    "SRET": 11 * 3600,
    "SRT": -3 * 3600,
    "SST": -11 * 3600,
    "SYOT": 3 * 3600,
    "T": -7 * 3600,
    "TAHT": -10 * 3600,
    "TFT": 5 * 3600,
    "TJT": 5 * 3600,
    "TKT": 13 * 3600,
    "TLT": 9 * 3600,
    "TMT": 5 * 3600,
    "TOST": 14 * 3600,
    "TOT": 13 * 3600,
    "TRT": 3 * 3600,
    "TVT": 12 * 3600,
    "U": -8 * 3600,
    "ULAST": 9 * 3600,
    "ULAT": 8 * 3600,
    "UTC": 0 * 3600,
    "UYST": -2 * 3600,
    "UYT": -3 * 3600,
    "UZT": 5 * 3600,
    "V": -9 * 3600,
    "VET": -4 * 3600,
    "VLAST": 11 * 3600,
    "VLAT": 10 * 3600,
    "VOST": 6 * 3600,
    "VUT": 11 * 3600,
    "W": -10 * 3600,
    "WAKT": 12 * 3600,
    "WARST": -3 * 3600,
    "WAST": 2 * 3600,
    "WAT": 1 * 3600,
    "WEST": 1 * 3600,
    "WET": 0 * 3600,
    "WFT": 12 * 3600,
    "WGST": -2 * 3600,
    "WGT": -3 * 3600,
    "WIB": 7 * 3600,
    "WIT": 9 * 3600,
    "WITA": 8 * 3600,
    "WST": 14 * 3600,
    "WT": 0 * 3600,
    "X": -11 * 3600,
    "Y": -12 * 3600,
    "YAKST": 10 * 3600,
    "YAKT": 9 * 3600,
    "YAPT": 10 * 3600,
    "YEKST": 6 * 3600,
    "YEKT": 5 * 3600,
    "Z": 0 * 3600,
}

# Fix float offsets. See https://github.com/goose3/goose3/issues/128.
TIMEZONE_INFO = {k: int(v) for k, v in TIMEZONE_INFO.items()}


class PublishDateExtractor(BaseExtractor):
    def extract(self):
        # check the opengraph and ReportageNewsArticle dictionary for the
        # publication date first.
        if "article:published_time" in self.article.opengraph:
            return self.article.opengraph["article:published_time"]
        if self.article.schema and "datePublished" in self.article.schema:
            date = self.article.schema["datePublished"]

            if type(date) == int or type(date) == float:
                t = None

                try:
                    # First except UNIX epoch in seconds.
                    t = datetime.fromtimestamp(date, tz=timezone.utc)
                except ValueError:
                    # In case if `date` was milliseconds we will get `ValueError: year N is out of range`.
                    # Then let's treat this `date` as UNIX epoch in milliseconds.
                    t = datetime.fromtimestamp(date / 1000, tz=timezone.utc)

                if t:
                    date = t.isoformat()
                else:
                    return None

            return date
        for known_meta_tag in self.config.known_publish_date_tags:
            # if this is a domain specific config and the current
            # article domain does not match the configured domain,
            # do not use the configured publish date pattern
            if known_meta_tag.domain and known_meta_tag.domain != self.article.domain:
                continue
            meta_tags = self.parser.get_elements_by_tag(
                self.article.doc, attr=known_meta_tag.attr, value=known_meta_tag.value, tag=known_meta_tag.tag
            )
            if meta_tags:
                if known_meta_tag.tag is None:
                    content = self.parser.get_attribute(meta_tags[0], known_meta_tag.content)
                    # if not content try to get from other publish date tags
                    if not content:
                        continue
                    if known_meta_tag.subcontent:
                        try:
                            subcontent = json.loads(content)

                            return subcontent[known_meta_tag.subcontent]
                        except (ValueError, KeyError):
                            return None
                    return content
                return meta_tags[0].text_content().strip()
        return None



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from copy import deepcopy

import lxml.html
from lxml import etree

from goose3.text import encode_value, get_encodings_from_content, inner_trim, smart_str
from goose3.utils import deprecated
from goose3.utils.constants import CAMEL_CASE_DEPRICATION


class Parser:
    @classmethod
    def xpath_re(cls, node, expression):
        regexp_namespace = "http://exslt.org/regular-expressions"
        items = node.xpath(expression, namespaces={"re": regexp_namespace})
        return items

    @classmethod
    def drop_tag(cls, nodes):
        if isinstance(nodes, list):
            for node in nodes:
                node.drop_tag()
        else:
            nodes.drop_tag()

    @classmethod
    def css_select(cls, node, selector):
        return node.cssselect(selector)

    @classmethod
    def fromstring(cls, html):
        encoding = get_encodings_from_content(html)
        encoding = encoding[0] if encoding else None
        if not encoding:
            html = encode_value(html)
            doc = lxml.html.fromstring(html)
        else:
            html = smart_str(html, encoding=encoding)
            parser = lxml.html.HTMLParser(encoding=encoding)
            doc = lxml.html.fromstring(html, parser=parser)
        return doc

    @classmethod
    def node_to_string(cls, node):
        return etree.tostring(node, encoding=str)

    @classmethod
    def replace_tag(cls, node, tag):
        node.tag = tag

    @classmethod
    def strip_tags(cls, node, *tags):
        etree.strip_tags(node, *tags)

    @classmethod
    def get_element_by_id(cls, node, idd):
        selector = f'//*[@id="{idd}"]'
        elems = node.xpath(selector)
        if elems:
            return elems[0]
        return None

    @classmethod
    def get_elements_by_tag(cls, node, tag=None, attr=None, value=None, childs=False):
        namespace = "http://exslt.org/regular-expressions"
        sel = tag or "*"
        selector = f"descendant-or-self::{sel}"
        if attr and value:
            selector = f'{selector}[re:test(@{attr}, "{value}", "i")]'
        elems = node.xpath(selector, namespaces={"re": namespace})
        # remove the root node
        # if we have a selection tag
        if node in elems and (tag or childs):
            elems.remove(node)
        return elems

    @classmethod
    def append_child(cls, node, child):
        node.append(child)

    @classmethod
    def child_nodes(cls, node):
        return list(node)

    @classmethod
    def child_nodes_with_text(cls, node):
        root = node
        # create the first text node
        # if we have some text in the node
        if root.text:
            elm = lxml.html.HtmlElement()
            elm.text = root.text
            elm.tag = "text"
            root.text = None
            root.insert(0, elm)
        # loop childs
        for _, elm in enumerate(list(root)):
            idx = root.index(elm)
            # don't process texts nodes
            if elm.tag == "text":
                continue
            # create a text node for tail
            if elm.tail:
                tmp = cls.create_element(tag="text", text=elm.tail, tail=None)
                root.insert(idx + 1, tmp)
        return list(root)

    @classmethod
    def text_to_para(cls, text):
        return cls.fromstring(text)

    @classmethod
    def get_children(cls, node):
        return node.getchildren()

    @classmethod
    def get_elements_by_tags(cls, node, tags):
        selector = ",".join(tags)
        elems = cls.css_select(node, selector)
        # remove the root node
        # if we have a selection tag
        if node in elems:
            elems.remove(node)
        return elems

    @classmethod
    def create_element(cls, tag="p", text=None, tail=None):
        elm = lxml.html.HtmlElement()
        elm.tag = tag
        elm.text = text
        elm.tail = tail
        return elm

    @classmethod
    def get_comments(cls, node):
        return node.xpath("//comment()")

    @classmethod
    def get_parent(cls, node):
        return node.getparent()

    @classmethod
    def remove(cls, node):
        parent = node.getparent()
        if parent is not None:
            if node.tail:
                prev = node.getprevious()
                if prev is None:
                    if not parent.text:
                        parent.text = ""
                    parent.text += " " + node.tail
                else:
                    if not prev.tail:
                        prev.tail = ""
                    prev.tail += " " + node.tail
            node.clear()
            parent.remove(node)

    @classmethod
    def get_tag(cls, node):
        return node.tag

    @classmethod
    def get_text(cls, node):
        return inner_trim(" ".join(node.itertext()).strip())

    @classmethod
    def previous_siblings(cls, node):
        nodes = []
        for _, val in enumerate(node.itersiblings(preceding=True)):
            nodes.append(val)
        return nodes

    @classmethod
    def previous_sibling(cls, node):
        nodes = []
        for idx, val in enumerate(node.itersiblings(preceding=True)):
            nodes.append(val)
            if idx == 0:
                break
        return nodes[0] if nodes else None

    @classmethod
    def next_sibling(cls, node):
        nodes = []
        for idx, val in enumerate(node.itersiblings(preceding=False)):
            nodes.append(val)
            if idx == 0:
                break
        return nodes[0] if nodes else None

    @classmethod
    def is_text_node(cls, node):
        return node.tag == "text"

    @classmethod
    def get_attribute(cls, node, attr=None):
        if attr:
            return node.attrib.get(attr, None)
        return attr

    @classmethod
    def del_attribute(cls, node, attr=None):
        if attr:
            _attr = node.attrib.get(attr, None)
            if _attr:
                del node.attrib[attr]

    @classmethod
    def set_attribute(cls, node, attr=None, value=None):
        if attr and value:
            node.set(attr, value)

    @classmethod
    def outer_html(cls, node):
        tmp = node
        if tmp.tail:
            tmp = deepcopy(tmp)
            tmp.tail = None
        return cls.node_to_string(tmp)

    # Aliases to previous names
    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use node_to_string instead")
    def nodeToString(cls, node):
        return cls.node_to_string(node)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use replace_tag instead")
    def replaceTag(cls, node, tag):
        return cls.replace_tag(node, tag)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use strip_tags instead")
    def stripTags(cls, node, *tags):
        return cls.strip_tags(node, tags)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use get_element_by_id instead")
    def getElementById(cls, node, idd):
        return cls.get_element_by_id(node, idd)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use get_elements_by_tag instead")
    def getElementsByTag(cls, node, tag=None, attr=None, value=None, childs=False):
        return cls.get_elements_by_tag(node, tag, attr, value, childs)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use append_child instead")
    def appendChild(cls, node, child):
        cls.append_child(node, child)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use child_nodes instead")
    def childNodes(cls, node):
        return cls.child_nodes(node)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use child_nodes_with_text instead")
    def childNodesWithText(cls, node):
        return cls.child_nodes_with_text(node)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use text_to_para instead")
    def textToPara(cls, text):
        return cls.text_to_para(text)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use get_children instead")
    def getChildren(cls, node):
        return cls.get_children(node)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use get_elements_by_tags instead")
    def getElementsByTags(cls, node, tags):
        return cls.get_elements_by_tags(node, tags)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use create_element instead")
    def createElement(cls, tag="p", text=None, tail=None):
        return cls.create_element(tag, text, tail)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use get_comments instead")
    def getComments(cls, node):
        return cls.get_comments(node)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use get_parent instead")
    def getParent(cls, node):
        return cls.get_parent(node)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use get_tag instead")
    def getTag(cls, node):
        return cls.get_tag(node)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use get_text instead")
    def getText(cls, node):
        return cls.get_text(node)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use previous_siblings instead")
    def previousSiblings(cls, node):
        return cls.previous_siblings(node)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use previous_sibling instead")
    def previousSibling(cls, node):
        return cls.previous_sibling(node)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use next_sibling instead")
    def nextSibling(cls, node):
        return cls.next_sibling(node)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use is_text_node instead")
    def isTextNode(cls, node):
        return cls.is_text_node(node)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use get_attribute instead")
    def getAttribute(cls, node, attr=None):
        return cls.get_attribute(node, attr)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use del_attribute instead")
    def delAttribute(cls, node, attr=None):
        cls.del_attribute(node, attr)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use set_attribute instead")
    def setAttribute(cls, node, attr=None, value=None):
        cls.set_attribute(node, attr, value)

    @classmethod
    @deprecated(f"Deprecated and to be removed in v{CAMEL_CASE_DEPRICATION}; use outer_html instead")
    def outerHtml(cls, node):
        return cls.outer_html(node)


class ParserSoup(Parser):
    @classmethod
    def fromstring(cls, html):
        from lxml.html import soupparser

        html = encode_value(html)
        doc = soupparser.fromstring(html)
        return doc



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import unittest

from goose3.configuration import Configuration
from goose3.image import Image, ImageDetails
from goose3.utils.images import ImageUtils

from .test_base import TestExtractionBase

CURRENT_PATH = os.path.dirname(os.path.abspath(__file__))


class ImageExtractionTests(TestExtractionBase):
    def getConfig(self):
        config = Configuration()
        config.enable_image_fetching = True
        return config

    def assert_top_image(self, fields, expected_value, result_image):
        # test if the result value
        # is an Goose Image instance
        msg = "Result value is not a Goose Image instance"
        self.assertTrue(isinstance(result_image, Image), msg=msg)

        # expected image
        expected_image = Image()
        for k, v in list(expected_value.items()):
            setattr(expected_image, f"_{k}", v)
        msg = "Expected value is not a Goose Image instance"
        self.assertTrue(isinstance(expected_image, Image), msg=msg)

        # check
        msg = "Returned Image is not the one expected"
        self.assertIn(result_image.src, expected_image.src, msg=msg)

        fields = vars(expected_image)
        for k, v in list(fields.items()):
            msg = "Returned Image attribute '%s' is not the one expected" % k
            self.assertEqual(getattr(expected_image, k), getattr(result_image, k), msg=msg)

    def test_basic_image(self):
        article = self.getArticle()
        fields = ["top_image"]
        self.runArticleAssertions(article=article, fields=fields)

    def test_base64_image(self):
        article = self.getArticle()
        fields = ["top_image"]
        self.runArticleAssertions(article=article, fields=fields)

    def _test_known_image_css(self, article):
        # check if we have an image in article.top_node
        images = self.parser.get_elements_by_tag(article.top_node, tag="img")
        self.assertEqual(len(images), 0)

        # we dont' have an image in article.top_node
        # check if the correct image was retrieved
        # using the known-image-css.txt
        fields = ["cleaned_text", "top_image"]
        self.runArticleAssertions(article=article, fields=fields)

    def test_known_image_name_parent(self):
        article = self.getArticle()
        self._test_known_image_css(article)

    def test_known_image_css_parent_class(self):
        article = self.getArticle()
        self._test_known_image_css(article)

    def test_known_image_css_parent_id(self):
        article = self.getArticle()
        self._test_known_image_css(article)

    def test_known_image_css_class(self):
        article = self.getArticle()
        self._test_known_image_css(article)

    def test_known_image_css_id(self):
        article = self.getArticle()
        self._test_known_image_css(article)

    def test_known_image_empty_src(self):
        "Tests that img tags for known image sources with empty src attributes are skipped."
        article = self.getArticle()
        self._test_known_image_css(article)

    def test_opengraph_tag(self):
        article = self.getArticle()
        self._test_known_image_css(article)

    def test_reportagenewsarticle(self):
        article = self.getArticle()
        self._test_known_image_css(article)


class ImageUtilsTests(unittest.TestCase):
    def setUp(self):
        self.path = f"{CURRENT_PATH}/data/images/50850547cc7310bc53e30e802c6318f1"
        self.expected_results = {"width": 476, "height": 317, "mime_type": "JPEG"}

    def test_utils_get_image_dimensions(self):
        image_detail = ImageUtils.get_image_dimensions(self.path)

        # test if we have an ImageDetails instance
        self.assertTrue(isinstance(image_detail, ImageDetails))

        # test image_detail attribute
        for k, v in list(self.expected_results.items()):
            self.assertEqual(getattr(image_detail, k), v)

    def test_detail(self):
        image_detail = ImageUtils.get_image_dimensions(self.path)

        # test if we have an ImageDetails instance
        self.assertTrue(isinstance(image_detail, ImageDetails))

        # test image_detail attribute
        for k, v in list(self.expected_results.items()):
            self.assertEqual(getattr(image_detail, k), v)

        # test image_detail get_ methode
        for k, v in list(self.expected_results.items()):
            attr = "get_%s" % k
            self.assertEqual(getattr(image_detail, attr)(), v)

        # test image_detail set_ methode
        expected_results = {"width": 10, "height": 10, "mime_type": "PNG"}

        for k, v in list(expected_results.items()):
            attr = "set_%s" % k
            getattr(image_detail, attr)(v)

        for k, v in list(expected_results.items()):
            attr = "get_%s" % k
            self.assertEqual(getattr(image_detail, attr)(), v)



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from goose3.extractors import BaseExtractor
from goose3.video import Video

VIDEOS_TAGS = ["iframe", "embed", "object", "video"]
VIDEO_PROVIDERS = ["youtube", "vimeo", "dailymotion", "kewego"]


class VideoExtractor(BaseExtractor):
    """Extracts a list of video from Article top node"""

    def __init__(self, config, article):
        super().__init__(config, article)
        self.candidates = []
        self.movies = []

    def get_embed_code(self, node):
        return "".join([line.strip() for line in self.parser.node_to_string(node).splitlines()])

    def get_embed_type(self, node):
        return self.parser.get_tag(node)

    def get_width(self, node):
        return self.parser.get_attribute(node, "width")

    def get_height(self, node):
        return self.parser.get_attribute(node, "height")

    def get_src(self, node):
        return self.parser.get_attribute(node, "src")

    @staticmethod
    def get_provider(src):
        if src:
            for provider in VIDEO_PROVIDERS:
                if provider in src:
                    return provider
        return None

    def get_video(self, node):
        """Create a video object from a video embed"""
        video = Video()
        video._embed_code = self.get_embed_code(node)
        video._embed_type = self.get_embed_type(node)
        video._width = self.get_width(node)
        video._height = self.get_height(node)
        video._src = self.get_src(node)
        video._provider = self.get_provider(video.src)
        return video

    def get_iframe_tag(self, node):
        return self.get_video(node)

    def get_embed_tag(self, node):
        # embed node may have an object node as parent
        # in this case we want to retrieve the object node
        # instead of the embed
        parent = self.parser.get_parent(node)
        if parent is not None:
            parent_tag = self.parser.get_tag(parent)
            if parent_tag == "object":
                return self.get_object_tag(node)
        return self.get_video(node)

    def get_object_tag(self, node):
        # test if object tag has en embed child
        # in this case we want to remove the embed from
        # the candidate list to avoid parsing it twice
        child_embed_tag = self.parser.get_elements_by_tag(node, "embed")
        if child_embed_tag and child_embed_tag[0] in self.candidates:
            self.candidates.remove(child_embed_tag[0])

        # get the object source
        # if wa don't have a src node don't coninue
        src_node = self.parser.get_elements_by_tag(node, tag="param", attr="name", value="movie")
        if not src_node:
            return None

        src = self.parser.get_attribute(src_node[0], "value")

        # check provider
        provider = self.get_provider(src)
        if not provider:
            return None

        video = self.get_video(node)
        video._provider = provider
        video._src = src
        return video

    def get_videos(self):
        # candidates node
        self.candidates = self.parser.get_elements_by_tags(self.article.top_node, VIDEOS_TAGS)

        # loop all candidates
        # and check if src attribute belongs to a video provider
        for candidate in self.candidates:
            tag = self.parser.get_tag(candidate)
            attr = f"get_{tag}_tag"
            if hasattr(self, attr):
                movie = getattr(self, attr)(candidate)
                if movie is not None and movie.provider is not None:
                    self.movies.append(movie)

        # append movies list to article
        return list(self.movies)



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import unittest

from goose3.text import StopWords, StopWordsKorean


class TestStopWordsBase(unittest.TestCase):
    def setUp(self):
        self.stopwords_class = StopWords
        self.lang = "en"

    def test_stop_words(self):
        # test that all the stopwords present in the paragraph are detected, and also
        # show that founds are valid

        text = "TextNode 1 - The Scala supported IDE is one of the few pain points of developers"
        text += " who want to start using Scala in their Java project."
        text += " On existing long term project developed by a team its hard to step in and introduce a new language"
        text += " that is not supported by the existing IDE."
        text += " On way to go about it is to hid the fact"
        text += " that you use Scala from the Java world by using one way dependency injection."
        text += " Still, if you wish to truly absorb Scala into your existing java environment"
        text += " then you'll soon introduced cross language dependencies."
        expected = [
            "the",
            "is",
            "one",
            "of",
            "the",
            "few",
            "of",
            "who",
            "want",
            "to",
            "using",
            "in",
            "their",
            "on",
            "by",
            "its",
            "to",
            "in",
            "and",
            "new",
            "that",
            "is",
            "not",
            "by",
            "the",
            "on",
            "way",
            "to",
            "go",
            "about",
            "it",
            "is",
            "to",
            "the",
            "that",
            "you",
            "use",
            "from",
            "the",
            "by",
            "using",
            "one",
            "way",
            "still",
            "if",
            "you",
            "wish",
            "to",
            "truly",
            "into",
            "your",
            "then",
            "soon",
        ]

        word_stats = self.stopwords_class(self.lang).get_stopword_count(text)

        self.assertEqual(word_stats.get_word_count(), 98)
        self.assertEqual(word_stats.get_stopword_count(), 53)
        self.assertCountEqual(word_stats.get_stop_words(), expected)
        self.assertListEqual(word_stats.get_stop_words(), expected)

        # check that founds are valid
        valid_stop_words = self.stopwords_class(self.lang)._stop_words
        for found_stop_word in word_stats.get_stop_words():
            self.assertIn(found_stop_word, valid_stop_words)


class TestStopWordsKorean(TestStopWordsBase):
    def setUp(self):
        self.stopwords_class = StopWordsKorean
        self.lang = "ko"

    def test_stop_words(self):
        text = "스칼라는 마틴 오더스키가 자바 제네릭 컴파일러를 개발하며 느꼈던 자바의 여러 가지 단점들을 근본적으로 수정하고,"
        text += " 추후 프로그램 언어 연구를 위한 연구 플랫폼으로 함께 사용하기 위하여 디자인한 언어이다."
        text += " 따라서 언뜻 보기에는 자바와 비슷해 보일지 모르나 여러가지 측면에서 더욱 발전된 형태를 가지고 있다."
        expected = [
            "라",
            "는",
            "가",
            "를",
            "며",
            "의",
            "여",
            "가",
            "을",
            "으로",
            "로",
            "하고",
            "고",
            "로",
            "를",
            "으로",
            "로",
            "께",
            "여",
            "이",
            "이다",
            "다",
            "라",
            "에",
            "에는",
            "는",
            "와",
            "나",
            "여",
            "가",
            "에",
            "에서",
            "를",
            "가",
            "고",
            "다",
        ]

        word_stats = self.stopwords_class(self.lang).get_stopword_count(text)

        self.assertEqual(word_stats.get_word_count(), 40)
        self.assertEqual(word_stats.get_stopword_count(), 36)
        self.assertCountEqual(word_stats.get_stop_words(), expected)
        self.assertListEqual(word_stats.get_stop_words(), expected)

        valid_stop_words = self.stopwords_class(self.lang)._stop_words
        for found_stop_word in word_stats.get_stop_words():
            self.assertIn(found_stop_word, valid_stop_words)



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json

from goose3.extractors import BaseExtractor

KNOWN_SCHEMA_TYPES = ("ReportageNewsArticle", "NewsArticle", "Article")


class SchemaExtractor(BaseExtractor):
    def extract(self):
        node = self.article.doc
        metas = self.parser.get_elements_by_tag(node, "script", attr="type", value="application/ld\\+json")
        for meta in metas:
            try:
                content = json.loads(meta.text_content())
                if isinstance(content, list):
                    for context in content:
                        if (
                            context["@context"] in ("https://schema.org", "http://schema.org")
                            and context["@type"] in KNOWN_SCHEMA_TYPES
                        ):
                            return context
                elif isinstance(content, dict):
                    if (
                        content["@context"] in ("https://schema.org", "http://schema.org")
                        and content["@type"] in KNOWN_SCHEMA_TYPES
                    ):
                        return content
            except (ValueError, KeyError):
                # If the contents are not proper JSON or a key we expect
                # to exist does not, continue to the next tag.
                continue
        return None



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import functools
import pkgutil
import time
import typing
import warnings
from typing import Union

from goose3.utils.constants import UINT64_T_MAX

KeyT = Union[str, bytes]


class FileHelper:
    @classmethod
    def load_resource_file(cls, filename):
        try:
            return pkgutil.get_data("goose3", filename).decode("utf-8")
        except OSError as exc:
            raise OSError(f"Couldn't open file '{filename}' - {exc}") from exc


class ParsingCandidate:
    def __init__(self, url_string: str, link_hash: str):
        self.url = url_string
        self.link_hash = link_hash


class RawHelper:
    @classmethod
    def get_parsing_candidate(cls, url: str, raw_html: str) -> ParsingCandidate:
        if isinstance(raw_html, str):
            raw_html = raw_html.encode("utf-8")
        link_hash = f"{fnv_1a(raw_html)}.{time.time()}"
        return ParsingCandidate(url, link_hash)


class URLHelper:
    @classmethod
    def get_parsing_candidate(cls, url_to_crawl: str) -> ParsingCandidate:
        # replace shebang is urls
        if "#!" in url_to_crawl:
            final_url = url_to_crawl.replace("#!", "?_escaped_fragment_=")
        else:
            final_url = url_to_crawl

        # url is only for calculating the link_hash
        url = final_url.encode("utf-8") if isinstance(final_url, str) else final_url

        link_hash = f"{fnv_1a(url)}.{time.time()}"
        return ParsingCandidate(final_url, link_hash)


class StringReplacement:
    def __init__(self, pattern, replace_with):
        self.pattern = pattern
        self.replace_with = replace_with

    def replace_all(self, string):
        if not string:
            return ""
        return string.replace(self.pattern, self.replace_with)


class ReplaceSequence:
    def __init__(self):
        self.replacements = []

    # @classmethod
    def create(self, first_pattern, replace_with=None):
        result = StringReplacement(first_pattern, replace_with or "")
        self.replacements.append(result)
        return self

    def append(self, pattern, replace_with=None):
        return self.create(pattern, replace_with)

    def replace_all(self, string):
        if not string:
            return ""

        mutated_string = string

        for itm in self.replacements:
            mutated_string = itm.replace_all(mutated_string)
        return mutated_string


def fnv_1a(key: KeyT, seed: int = 0) -> str:
    """Pure python implementation of the 64 bit fnv-1a hash
    Args:
        key (str): The element to be hashed
        seed (int): Add a seed to the initial starting point (0 means no seed)
    Returns:
        str: 64-bit hashed representation of key
    Note:
        Uses the lower 64 bits when overflows occur"""
    max64mod = UINT64_T_MAX + 1
    hval = (14695981039346656037 + (31 * seed)) % max64mod
    fnv_64_prime = 1099511628211
    tmp = list(key) if not isinstance(key, str) else list(map(ord, key))
    for t_str in tmp:
        hval ^= t_str
        hval *= fnv_64_prime
        hval %= max64mod
    # Remove the leading "0x".
    return hex(hval)[2:]


def deprecated(message: str = "") -> typing.Callable:
    """A simplistic decorator to mark functions as deprecated. The function
    will pass a message to the user on the first use of the function

    Args:
        message (str): The message to display if the function is deprecated
    """

    def decorator_wrapper(func):
        @functools.wraps(func)
        def function_wrapper(*args, **kwargs):
            func_name = func.__name__
            if func_name not in function_wrapper.deprecated_items:
                msg = f"Function {func.__name__} is now deprecated! {message}"
                warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
                function_wrapper.deprecated_items.add(func_name)

            return func(*args, **kwargs)

        # set this up the first time the decorator is called
        function_wrapper.deprecated_items = set()

        return function_wrapper

    return decorator_wrapper



"""
This is a python port of "Goose" orignialy licensed to Gravity.com
under one or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.

Python port was written by Xavier Grangier for Recrutae

Gravity.com licenses this file
to you under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from goose3 import Configuration, PublishDatePattern

from .test_base import TestExtractionBase


class TestPublishDate(TestExtractionBase):
    def test_publish_date(self):
        article = self.getArticle()
        self.runArticleAssertions(article=article, fields=["publish_date", "publish_datetime_utc"])

    def test_publish_date_rnews(self):
        article = self.getArticle()
        self.runArticleAssertions(article=article, fields=["publish_date", "publish_datetime_utc"])

    def test_publish_date_article(self):
        article = self.getArticle()
        self.runArticleAssertions(article=article, fields=["publish_date", "publish_datetime_utc"])

    def test_publish_date_schema(self):
        article = self.getArticle()
        self.runArticleAssertions(article=article, fields=["publish_date", "publish_datetime_utc"])

    def test_publish_date_unix_milliseconds(self):
        article = self.getArticle()
        self.runArticleAssertions(article=article, fields=["publish_date", "publish_datetime_utc"])

    def test_publish_date_unix_seconds(self):
        article = self.getArticle()
        self.runArticleAssertions(article=article, fields=["publish_date", "publish_datetime_utc"])

    def test_publish_date_parsely_page(self):
        article = self.getArticle()
        self.runArticleAssertions(article=article, fields=["publish_date", "publish_datetime_utc"])

    def test_publish_date_tag(self):
        config = {"known_publish_date_tags": {"attribute": "class", "value": "pubdate", "tag": "div"}}
        article = self.getArticle(config)
        self.runArticleAssertions(article=article, fields=["publish_date", "publish_datetime_utc"])

    def test_publish_date_config(self):
        # Test with configuring the pubdate with a dict and no domain filter
        config0 = {"known_publish_date_tags": {"attribute": "name", "value": "super-rare-date-tag", "content": "value"}}
        article = self.getArticle(config_=config0)
        self.runArticleAssertions(article=article, fields=["publish_date", "publish_datetime_utc"])

        # Test with configuring the pubdate with a PublishDatePattern and no domain filter
        config1 = {
            "known_publish_date_tags": PublishDatePattern(attr="name", value="super-rare-date-tag", content="value")
        }
        article = self.getArticle(config_=config1)
        self.runArticleAssertions(article=article, fields=["publish_date", "publish_datetime_utc"])

        # Test with configuring the incorrect domain filter
        config2 = {
            "known_publish_date_tags": PublishDatePattern(
                attr="name", value="super-rare-date-tag", content="value", domain="incorrect.com"
            )
        }
        article = self.getArticle(config_=config2)
        assert not article.publish_date
        assert not article.publish_datetime_utc

        # Test with configuring the correct domain filter
        config3 = {
            "known_publish_date_tags": PublishDatePattern(
                attr="name", value="super-rare-date-tag", content="value", domain="example.com"
            )
        }
        article = self.getArticle(config_=config3)
        self.runArticleAssertions(article=article, fields=["publish_date", "publish_datetime_utc"])
